package be.iminds.ilabt.jfed.highlevel.tasks;

import be.iminds.ilabt.jfed.fedmon.webapi.client.FedmonWebApiClient;
import be.iminds.ilabt.jfed.fedmon.webapi.client.LastResultFilter;
import be.iminds.ilabt.jfed.fedmon.webapi.client.TestInstanceFilter;
import be.iminds.ilabt.jfed.highlevel.controller.Task;
import be.iminds.ilabt.jfed.highlevel.controller.TaskExecution;

import be.iminds.ilabt.jfed.lowlevel.*;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.*;
import be.iminds.ilabt.jfed.lowlevel.connection.HttpConnection;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnectionProvider;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.util.JavaFXLogger;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.Collections;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;

/**
 * User: twalcari
 * Date: 6/17/2015
 * Time: 11:30
 */
public class FedmonResultsTask extends Task {
    private static final Logger LOG = LoggerFactory.getLogger(FedmonResultsTask.class);

    private static final String FEDMON_URN = "urn:publicid:IDN+flsmonitor.fed4fire.eu+authority+cm";
    //input parameters
    @Nullable public final String testbedFilter;
    @Nullable public final String testbedUrnFilter;
    @Nullable public final String testDefinitionFilter;

    //results
    public List<Result> results = null;

    //environment parameters
    @Nonnull private final JavaFXLogger logger;
    @Nonnull private final GeniUserProvider geniUserProvider;
    @Nonnull private final JFedPreferences jFedPreferences;
    @Nonnull private final TestbedInfoSource testbedInfoSource;
    @Nonnull private final JFedConnectionProvider connectionProvider;
    @Nonnull private final FedmonWebApiClient fedmonWebApiClient;


    public FedmonResultsTask(@Nullable String testbedFilter,
                             @Nullable String testbedUrnFilter,
                             @Nullable String testDefinitionFilter,
                             @Nonnull JavaFXLogger logger,
                             @Nonnull GeniUserProvider geniUserProvider,
                             @Nonnull JFedPreferences jFedPreferences,
                             @Nonnull TestbedInfoSource testbedInfoSource,
                             @Nonnull JFedConnectionProvider connectionProvider,
                             @Nonnull FedmonWebApiClient fedmonWebApiClient) {
        super("Get Fedmon Results");
        this.testbedFilter = testbedFilter;
        this.testbedUrnFilter = testbedUrnFilter;
        this.testDefinitionFilter = testDefinitionFilter;
        this.logger = logger;
        this.geniUserProvider = geniUserProvider;
        this.jFedPreferences = jFedPreferences;
        this.testbedInfoSource = testbedInfoSource;
        this.connectionProvider = connectionProvider;
        this.fedmonWebApiClient = fedmonWebApiClient;
    }

    @Override
    protected void doTask(@Nonnull TaskExecution taskExecution) throws JFedException, InterruptedException {
        Server fedmonAuth = testbedInfoSource.getByUrnExact(FEDMON_URN);

        TestInstanceFilter testInstanceFilter = new TestInstanceFilter(
                null, null,
                testDefinitionFilter,
                null, null, null,
                testbedFilter, testbedUrnFilter,
                Optional.of(false));
        LastResultFilter filter = new LastResultFilter(testInstanceFilter,
                true, java.util.Optional.empty(), null);
        try {
            this.results = fedmonWebApiClient.search(filter);
            LOG.debug("Found {} fedmon results", this.results.size());
        } catch (FedmonWebApiClient.FedmonWebApiClientException e) {
            this.results = Collections.emptyList();
            throw new JFedException("Failed to fetch fedmon results", e);
        }
    }

    public List<Result> getResults() {
        return results;
    }

    public List<Result> getSuccessResults() {
        if (results != null) {
            return results.stream()
                    .filter(r -> r.getSummaryStatus() == Result.ResultStatus.SUCCESS)
                    .collect(Collectors.toList());
        } else
            return null;
    }
}
